/*
**  SLOPE_REV: C-XSC classes for one- and multidimensional (interval)
**             slope arithmetic in reverse mode
**
**  Copyright (C) 1999-2001 M. Braeuer, W. Hofschuster, W. Kraemer
**                          Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

////////////////////////////////////////////////////////////////////
//
// s_reverse.h       Headerfile fuer mehrdimensionale 
// Intervallsteigungsarithmetik durch Rueckwaertsrechnung
//
/////////////////////////////////////////////////////////////////////

#ifndef __SREVERSE_H
#define __SREVERSE_H 

//cxsc headers
#include "interval.hpp"
#include "imath.hpp"
#include "intvector.hpp"
#include "ivector.hpp"
#include "imatrix.hpp"
#include "rvector.hpp"

//gnu g++ headers
#include <iostream>
#include <stdlib.h>

//Headerfile fuer Einbinden der Hilfsfunktionen
#include "helpfuncs.h"

namespace cxsc {

enum{
// zum Ueberladen der Verzweigungs- und Standardfunktionen

__sqrt,
__sqrt_n,
__sqr,

__sin,
__cos,        
__tan,
__cot,

__asin,
__acos,
__atan,
__acot,

__exp,
__ln,

__sinh,
__cosh,
__tanh,
__coth,

__asinh,
__acosh,
__atanh,
__acoth,

__power_n,

__abs,
__max,
__chi,

__x = 50,
__c,

__unminus,

__plus,
__minus,
__mul,
__div,

__erg
};

/*----------------------------------------------------------------------
Compiler: g++/gcc version 2.7.2.1
------------------------------------------------------------------------*/


class STAPE;

//////////////////////////////////////////////////////////////////////////
//
//Definitionen der Klassen
//
//////////////////////////////////////////////////////////////////////////
/*-----------------------------------------------------------------------

class rslope  Klasse fuer Objekte zur automatische Berechnung mehrdim-
              ensionaler Intervallsteigungen, werden auf Tape geschrieben
              
  Elemente: 
            STAPE* tapezeiger .... Zeiger auf aktuelles Tape
            int k ................ Nummer der Operation
            interval  f_x ........ 1.Intervalkomponente (Fkt.wert)
            interval  f_c ........ 2.Intervalkomponente (Fkt.wert)
            int op ............... im k-ten Schritt durchzufuehrende Op.
            int l ................ erster Vorgaenger
            int r ................ zweiter Vorgaenger
            int bed .............. Kontrolleintrag fuer Bedingung

  Elementfunktionen:

            siehe Implementation 

Implementationsfile: s_reverse.cpp

-----------------------------------------------------------------------*/

class rslope{
 //friend class STAPE; waere besser

 private:
  STAPE* tapezeiger; 
  int k ;              
  interval f_x;
  interval f_c;   
  int  op ;         
  int l;           
  int r; 
  int bed; 

 public:
  
  rslope();
  rslope(STAPE& );
  rslope(STAPE& , const interval& );
  rslope(STAPE& , const real& );
  rslope(STAPE& , const interval& , const interval& );
  rslope(const rslope& );

  rslope& operator=(const rslope& );

  friend void print_rslope(const rslope& ); 

  interval get_f_x(){return f_x;};
  interval get_f_c(){return f_c;};
  int get_k(){return k;};
  int get_op(){return op;};
  int get_l(){return l;};
  int get_r(){return r;};
  int get_bed(){return bed;};

  friend void rslope_anfuegen(STAPE& working_tape, 
			      rslope& s, 
			      const interval& fx,
			      const interval& fc,
			      int op, 
			      int l, 
			      int r, 
			      int bed );

  friend rslope result(const rslope& u); 

  friend rslope operator-(const rslope& u);

  friend rslope operator+(const rslope& u, const rslope& v);
  friend rslope operator-(const rslope& u, const rslope& v);
  friend rslope operator*(const rslope& u, const rslope& v);   
  friend rslope operator/(const rslope& u, const rslope& v);

  friend rslope operator+(const interval& xc, const rslope& u);
  friend rslope operator-(const interval& xc, const rslope& u);
  friend rslope operator*(const interval& xc, const rslope& u);
  friend rslope operator/(const interval& xc, const rslope& u);

  friend rslope operator+(const rslope& u, const interval& xc);
  friend rslope operator-(const rslope& u, const interval& xc);
  friend rslope operator*(const rslope& u, const interval& xc);
  friend rslope operator/(const rslope& u, const interval& xc);

  friend rslope operator+(real xc, const rslope& u);
  friend rslope operator-(real xc, const rslope& u);
  friend rslope operator*(real xc, const rslope& u);
  friend rslope operator/(real xc, const rslope& u);

  friend rslope operator+(const rslope& u, real xc);
  friend rslope operator-(const rslope& u, real xc);
  friend rslope operator*(const rslope& u, real xc);
  friend rslope operator/(const rslope& u, real xc);

  friend rslope sqrt(const rslope& u);
  friend rslope sqrt(const rslope& u, int n);
  friend rslope sqr(const rslope& u);
  friend rslope power(const rslope& u, int n);
  friend rslope pow(const rslope& u, const interval& alpha);

  friend rslope sin(const rslope& u);
  friend rslope cos(const rslope& u);
  friend rslope tan(const rslope& u);
  friend rslope cot(const rslope& u);

  friend rslope asin(const rslope& u);
  friend rslope acos(const rslope& u);
  friend rslope atan(const rslope& u);
  friend rslope acot(const rslope& u);

  friend rslope exp(const rslope& u);
  friend rslope ln(const rslope& u);

  friend rslope sinh(const rslope& u);
  friend rslope cosh(const rslope& u);
  friend rslope tanh(const rslope& u);
  friend rslope coth(const rslope& u);

  friend rslope asinh(const rslope& u);
  friend rslope acosh(const rslope& u);
  friend rslope atanh(const rslope& u);
  friend rslope acoth(const rslope& u);

  friend rslope abs(const rslope& ); 
  friend rslope max(const rslope&, const rslope& ); 
  friend rslope min(const rslope&, const rslope& );   
  friend rslope chi(const rslope&, const rslope&, const rslope& ); 

};


/*-----------------------------------------------------------------------

 class rslope_vec: Vektor mit Elementen der Klasse rslope
                   zum Initialisieren der Unabhaengigen und Abhaeng.

-----------------------------------------------------------------------*/
class rslope_vec{
 private: 
  int dim;
  int lb; //=1 !
  int ub;
  rslope* rslope_vector; 
     
 public:
  rslope_vec(int n=1);

  rslope_vec(const rslope_vec& );
  ~rslope_vec();
  rslope_vec& operator=(const rslope_vec& );
  rslope& operator[](int ); 
  rslope& operator[](int ) const; 
  int size();

  friend int Lb(const rslope_vec& );
  friend int Ub(const rslope_vec& );
};


/*-----------------------------------------------------------------------

class list_rslope: fuer liste mit Elementen der Klasse rslope

-----------------------------------------------------------------------*/
class list_rslope{
 class list_rslope_element{
  public:
  
  list_rslope_element* suc;
  int count; //fuer Indexzugriff
  rslope element;
 };
 typedef list_rslope_element* node;
 node head;
 node back;

 public:
  list_rslope();
  list_rslope(const rslope& );
  ~list_rslope();

  rslope& operator[](int );
  rslope& operator[](int ) const;

  void push__back(const rslope& e);  
  void clean_up_list_rslope();
  void print_list_rslope();
  int size();
};

/*-----------------------------------------------------------------------

class STAPE: beinhaltet Liste vom Type list_rslope zum Aufzeichnen der
             Codeliste

-----------------------------------------------------------------------*/
class STAPE{         

 private:
  list_rslope tape; //Liste von Objekten des Typs rslope    
  int laenge_tape;  //Laenge des Tapes/Liste

  int stop_writing; // falls 1: es wird kein rslope mehr angefuegt

  int counter;      //Zaehler fuer _erg's
  intvector y_comp; //Stellen in tape, wo _erg geschrieben wird
 
  int dimension_x; //Dimension Unabhaengige
  int dimension_y; //Dimension Abhaengige

  static int writing; //Flag fuer Ausschliessen von Verschachtelungen
                      //ist 1 falls irgendwo ein Tape geschrieben wird,
                      //sonst 0

 public:

  friend void init_tape(STAPE&, 
		        rslope_vec& , const ivector& ,
	                rslope_vec& );


  friend void init_tape(STAPE& , 
		        rslope_vec& , const  ivector& , const ivector& ,
		        rslope_vec& );

  friend void init_tape(STAPE& ,
		        rslope_vec& , const  ivector& , const rvector&,
		        rslope_vec& );


  friend void stop_writing_tape(STAPE& working_tape);

  friend void rslopeVar(STAPE& working_tape,  rslope_vec& ein_vec,
                        const ivector& x, const ivector& c);

  friend void rslopeVar(STAPE& working_tape, rslope_vec& ein_vec,
                        const ivector& x, const rvector& c);

  friend void rslopeVar_mid(STAPE& working_tape, rslope_vec& ein_vec,
                           const ivector& x);

  friend void rslope_anfuegen(STAPE& working_tape, 
			      rslope& s,  
			      const interval& fx,
			      const interval& fc,
			      int op, 
			      int l, 
			      int r, 
			      int bed );
  
  int get_laenge_tape();         //gibt Laenge+1 des Tapes aus       
  void set_laenge_tape(int n=0); //setzt Laenge des Tapes auf 0

  void set_stop_writing(const int& n=1);
  int  get_stop_writing();

  int get_counter();
  void set_counter(int );

  int get_y_comp(int );
  void set_y_comp(int ,int ); 

  int get_size(); //gibt wirkliche Laenge der Liste zurueck

  int get_dimension_x();
  int get_dimension_y();

  friend void print_tape(STAPE& working_tape);

  friend void clean_up(STAPE& working_tape);


  friend void reverse_sweep(STAPE& wt, const ivector& slopes, int y_c);
  friend void reverse_sweep(STAPE& wt, imatrix& slopematrix);
};


//verschiedene Funktionen fuer forward-sweep


void forward_sweep(STAPE& working_tape, 
		   rslope (*f)(STAPE& working_tape, const rslope_vec& x),
                   const ivector& x_in,
                   interval& fx_val,
                   interval& fc_val);

void forward_sweep(STAPE& working_tape, 
		   rslope (*f)(STAPE& working_tape, const rslope_vec& x),
                   const ivector& x_in, const ivector& c_in,
                   interval& fx_val,
                   interval& fc_val);

void forward_sweep(STAPE& working_tape, 
		   rslope (*f)(STAPE& working_tape, const rslope_vec& x),
                   const ivector& x_in, rvector& c_in,
                   interval& fx_val,
                   interval& fc_val);

 //vektorwertige Funktionen

void forward_sweep(STAPE& working_tape, 
		   rslope_vec (*f)(STAPE& working_tape, const rslope_vec& x),
                   const ivector& x_in,
                   ivector& fx_val,
                   ivector& fc_val);

void forward_sweep(STAPE& working_tape, 
		   rslope_vec (*f)(STAPE& working_tape, const rslope_vec& x),
                   const ivector& x_in, const ivector& c_in,
                   ivector& fx_val,
                   ivector& fc_val);

void forward_sweep(STAPE& working_tape, 
		   rslope_vec (*f)(STAPE& working_tape, const rslope_vec& x),
                   const ivector& x_in, rvector& c_in,
                   ivector& fx_val,
                   ivector& fc_val);

} //namespace cxsc
#endif
